package fcr;

###############################################################################
#							FC_RECUR.PL
#
# Version: 1.1.1
#
# Last Mofified: 29 MAR 1999
#
# This program was written by Michael E. Schechter (mschechter@earthlink.net)
# to provide for recurring events in the FrameCal Perl application.  Portions
# of this series of subroutines are inspired by Date::Manip.pm, written by
# Sullivan Beck.  Date::Manip.pm provides for recurring times, where they
# are defined by a recurrence string of Y:M:W:D:H:M:S, and many other things.
# However, the module is 253 KB, and provides far more than is needed by
# FrameCal.  So, after studying the module, the basic idea was modified to
# provide only what was needed by FrameCal, and thus, the file is only
# 6 KB when comments have been stripped out.
#
# You may copy this file under the terms of the GNU General Public License.
###############################################################################

###############################################################################
# Subroutine:  get_dates()
#
#   Parameters:
#		$recur = A string defining the recurrence.
#		$datea = the start date as yyyymmdd.
#		$dateb = the "through" date as yyyymmdd.
#
#   Usage:
#		@dates = &get_dates( $recur, $datea, $dateb );
#
#   Output:
#		@dates = an array of dates in the form yyyymmdd that match the
#				 recurrence string and fall between the dates passed.
#
#	Notes:
#		This subroutine takes two dates and a recurrence string.  The string
#		defines the recurrences in the following manner:
#
#		Each string will contain four fields, that match Y:M:W:D.  Each string
#		may contain one asterisk, and is interpreted with the following
#		examples:
#					   0:2:0:0 = Every two months
#					   0:0:2*1 = Every other Monday
#					 0:0:1*2,4 = Every Tuesday and Thursday
#				0:1*2,-1:1,3,5 = Every month on the second and last Monday,
#								 Wednesday, and Friday
#					   2:0*1:0 = Every other year on the first Sunday
###############################################################################

sub get_dates {
	my( $recur, $datea, $dateb ) = @_;
	my( $rec_a, $rec_b, @rec_a, @rec_b, @dates ) = ();
	( $rec_a, $rec_b ) = (split(/\*/, $recur));
	@rec_a = split(/:/, $rec_a);
	@rec_b = split(/:/, $rec_b);
	if( (scalar(@rec_a) + scalar(@rec_b)) == 4 ) {
		if( $#rec_a == 3 ) {
			if( $rec_a =~ /^(?:0:){0,3}[1,2,3,4](?::0){0,3}$/o ) {
				@dates = &date_recur( $recur, $datea, $dateb );
			}
		} elsif( $#rec_a == 2 ) {
			my( @dateas, $date, $tempdate ) = ();
			my( @days ) = &return_list( $rec_b );
			my( $interval ) = pop( @rec_a );
			if( ($interval =~ /(?:1|2|3|4)/o) && (&is_val_range(\@days,0,6)) ) {
				foreach $day (@days) {
					$tempdate = &get_first( $datea, $dateb, $day );
					push( @dateas, $tempdate ) if( $tempdate );
				}
				foreach $date (@dateas) {
					push( @dates, &date_recur( "0:0:$interval:0", $date, $dateb ) );
				}
				@dates = sort { $a <=> $b } ( @dates );
			}
		} elsif( $#rec_a == 1 ) {
			my( $delta, @tmp, @w, @d, $temp ) = ();
			@tmp = @rec_a;
			@w = &return_list( shift(@rec_b) );
			@d = &return_list( shift(@rec_b) );
			if( ($rec_a =~ /^(?:0:[1,2,3,4,6]|1:0)$/o) && (&is_val_range(\@w,-1,3))
				&& (&is_val_range(\@d,0,6)) ) {
				push( @tmp, 0 ) while( $#tmp < 3 );
				$delta = join(":", @tmp );
				@tmp = &date_recur( $delta, $datea, $dateb );
				@dates = &date_recur_wom( \@tmp, \@w, \@d );
				$temp = $dates[ 0 ];
				while( ($temp < $datea) && scalar(@dates) ) {
					$temp = shift( @dates );
					$temp = $dates[ 0 ];
				}
				$temp = $dates[ -1 ];
				while( ($temp > $dateb) && scalar(@dates) ) {
					$temp = pop( @dates );
					$temp = $dates[ -1 ];
				}
			}
		}
	}
	return( @dates );
}

###############################################################################
# Subroutine:  date_recur()
#
#   Parameters:
#		$recur = A string defining the recurrence.
#		$datea = the start date as yyyymmdd.
#		$dateb = the "through" date as yyyymmdd.
#
#   Usage:
#		@dates = &date_recur( $recur, $datea, $dateb );
#
#   Output:
#		@dates = an array of dates in the form yyyymmdd that match the
#				 recurrence string and fall between the dates passed.
#
#	Note:
#		This subroutine differs from &get_dates above in the case that
#		it only handles recurrences of intervals of years, or months,
#		or weeks, or days, but never a combination of any of them.  There
#		will be no lists supplied in this recurrence string.
###############################################################################

sub date_recur {
	my( $recur, $datea, $dateb ) = @_;
	my( $y, $m, $w, $d, @dates, $jda, $jdb ) = ();
	push( @dates, $datea );
	($y, $m, $w, $d) = split(/:/, $recur);
	if( $w || $d ) {
		$jda = &julian_date( &date_split($datea) );
		$jdb = &julian_date( &date_split($dateb) );
	}
	if( $y != 0 ) {
		my( $rec_int ) = $y * 10000;
		$datea += $rec_int;
		while( $datea <= $dateb ) {
			push( @dates, $datea ) if( &is_val_date($datea) );
			$datea += $rec_int;
		}
	} elsif( $m != 0 ) {
		my( $ym, $mm, $dm ) = &date_split( $datea );
		$mm += $m;
		if( $mm > 12 ) {
			$ym++;
			$mm -= 12;
		}
		$datea = &date_join( $ym, $mm, $dm );
		while( $datea <= $dateb ) {
			push( @dates, $datea ) if( &is_val_date($datea) );
			$mm += $m;
			if( $mm > 12 ) {
				$ym++;
				$mm -= 12;
			}
			$datea = &date_join( $ym, $mm, $dm );
		}
	} elsif( $w != 0 ) {
		my( $rec_int ) = $w * 7;
		$jda += $rec_int;
		while( $jda <= $jdb ) {
			push( @dates, &date_join( &jdn_to_date($jda) ) );
			$jda += $rec_int;
		}
	} else {
		$jda += $d;
		while( $jda <= $jdb ) {
			push( @dates, &date_join( &jdn_to_date($jda) ) );
			$jda += $d;
		}
	}
	return( @dates );
}

###############################################################################
# Subroutine:  date_recur_wom()
#
#   Parameters:
#		$y = a reference to an array of years.
#		$w = a reference to an array of weeks.
#		$d = a reference to an array of days.
#
#   Usage:
#		@dates = &date_recur_wom( \@years, \@weeks, \@days );
#
#   Output:
#		@dates = an array of dates in the form yyyymmdd that fall in the
#				 years specified, and are in the list of weeks and days
#				 passed.
#
#	Note:
#		This subroutine takes an array of years, weeks, and days.  The years
#		array is actually an array of dates, which are split into a new array
#		of years and months that fall within the original recurrence that
#		invoked this subroutine.  The weeks list specifies items such as
#		first week, third week, last week, etc.  The days list specifies the
#		days of the week that are to be found, such as Sunday, Tuesday, etc.
###############################################################################

sub date_recur_wom {
	my( $y, $w, $d ) = @_;
	my( @y ) = @$y;
	my( @w ) = @$w;
	my( @d ) = @$d;
	my( @m, @dates, @tmpa, @tmpb, $ya, $ma, $da, $db, $wa ) = ();
	my( $datea, $dateb, $datec, %final ) = ();
	foreach $ya ( @y ) {
		@tmpa = &date_split( $ya );
		push( @tmpb, $tmpa[ 0 ] );
		push( @m, $tmpa[ 1 ] );
	}
	@y = @tmpb;
	foreach $ya ( @y ) {
		( @tmpa, @tmpb, %final ) = ();
		$ma = shift( @m );
		$datea = &date_join( $ya, $ma, 1 );
		$db = &main::days_in_month( $ya, $ma );
		$dateb = &date_join( $ya, $ma, $db );
		foreach $da ( @d ) {
			$datec = &get_first( $datea, $dateb, $da );
			@tmpa = &date_recur( '0:0:1:0', $datec, $dateb );
			foreach $wa ( @w ) {
				$final{ $tmpa[ $wa ] } = 1;
			}
		}
		@tmpb = sort { $a <=> $b } ( keys( %final ) );
		push( @dates, @tmpb );
	}
	return( @dates );
}

###############################################################################
# Subroutine:  date_join()
#
#   Parameters:
#		 $year = the year.
#		$month = the month.
#		  $day = the day.
#
#   Usage:
#		$date = &date_join( $year, $month, $day );
#
#   Output:
#		$date = a date in the format of yyyymmdd.
###############################################################################

sub date_join {
	my( $year, $month, $day ) = @_;
	$month = "0$month" if( length($month) == 1 );
	$day = "0$day" if( length($day) == 1 );
	return( "$year$month$day" );
}

###############################################################################
# Subroutine:  date_split()
#
#   Parameters:
#		 $date = a date in the form of yyyymmdd.
#
#   Usage:
#		( $year, $month, $day ) = &date_split( $date );
#
#   Output:
#		 $year = the year of the date passed.
#		$month = the month of the date passed.
#		  $day = the day of the date passed.
###############################################################################

sub date_split {
	my( $date ) = @_;
	my( $year, $month, $day ) = ();
	$date =~ /^(\d{4})(\d{2})(\d{2})$/o;
	( $year, $month, $day ) = ( $1, $2, $3 );
	return( $year, $month, $day );
}

###############################################################################
# Subroutine:  return_list()
#
#   Parameters:
#		 $list = a comma separated list of numbers.
#
#   Usage:
#		@list = &return_list( $list );
#
#   Output:
#		@list = an array of the numbers in the comma separated list.
###############################################################################

sub return_list {
	my( $list ) = @_;
	my( @list ) = ();
	if( $list =~ /,/o ) {
		@list = split(/,/, $list);
	} else {
		push( @list, $list );
	}
	return( @list );
}

###############################################################################
# Subroutine:  get_first()
#
#   Parameters:
#		$datea = the start date.
#		$dateb = the end date.
#		  $dow = the day of the week we are trying to find.
#
#   Usage:
#		$date = &get_first( $datea, $dateb, $dow );
#
#   Output:
#		$date = the first instance of the $dow that falls between $datea and
#				$dateb.
###############################################################################

sub get_first {
	my( $datea, $dateb, $dow ) = @_;
	my( $jda, $adder, $tempdate ) = ();
	my( $ya, $ma, $da ) = &date_split( $datea );
	my( $datea_dow ) = &main::day_of_week_num( $ya, $ma, $da );
	$jda = &julian_date( $ya, $ma, $da );
	if( $datea_dow <= $dow ) {
		$adder = $dow - $datea_dow;
	} else {
		$adder = (7 + $dow) - $datea_dow;
	}
	$jda += $adder;
	$tempdate = &date_join( &jdn_to_date($jda) );
	unless( $tempdate <= $dateb ) {
		$tempdate = "";
	}
	return( $tempdate );
}

###############################################################################
# Subroutine:  julian_date()
#
#   Parameters:
#		 $year = the year.
#		$month = the month.
#		  $day = the day.
#
#   Usage:
#		$jdn = &julian_date( $year, $month, $day );
#
#   Output:
#		$jdn = the julain day number for the date specified by the year,
#			   month, and day passed into the subroutine.
###############################################################################

sub julian_date {
	my( $year, $month, $day ) = @_;
	my( $a, $b, $c, $jdn ) = ();
	$a = int((14 - $month)/12);
	$b = $year + 4800 - $a;
	$c = $month + (12 * $a) - 3;
	$jdn = $day+int((153*$c+2)/5)+(365*$b)+int($b/4)-int($b/100)+int($b/400)-32045;
	return( $jdn );
}

###############################################################################
# Subroutine:  jdn_to_date()
#
#   Parameters:
#		$jdn = a julian day number.
#
#   Usage:
#		( $year, $month, $day ) = &jdn_to_date( $jdn );
#
#   Output:
#		( $year, $month, $day ) = the year, month, and day that correspond to
#								  the passed in julian day number.
###############################################################################

sub jdn_to_date {
	my( $jdn ) = @_;
	my( $a, $b, $c, $d, $e, $m, $year, $month, $day ) = ();
	$a = $jdn + 32045;
	$b = int( (4*($a+36524))/146097 ) - 1;
	$c = $a - int( (146097*$b)/4 );
	$d = int( (4*($c+365))/1461 ) - 1;
	$e = $c - int( (1461*$d)/4 );
	$m = int( (5*($e-1)+2)/153 );
	$day = $e - int( (153*$m+2)/5 );
	$month = $m + 3 - (12 * int($m/10));
	$year = (100*$b) + $d - 4800 + int($m/10);
	return( $year, $month, $day );
}

###############################################################################
# Subroutine:  is_val_date()
#
#   Parameters:
#		$date = a date in the form of yyyymmdd.
#
#   Usage:
#		( &is_val_date( $date ) );
#
#   Output:
#		This is a boolean value return, basically.  The subroutine returns 1,
#		or true, if the date is a valid date in the Gregorian calendar, 0, or
#		false, otherwise.
###############################################################################

sub is_val_date {
	my( $date ) = @_;
	my( $year, $month, $day ) = &date_split( $date );
	my( $bool ) = 1;
	my( $num_days ) = &main::days_in_month( $year, $month );
	if( ($month < 1) || ($month > 12) ) {
		$bool = 0;
	} elsif( ($day < 1) || ($day > $num_days) ) {
		$bool = 0;
	} elsif( $year < 1582 ) {
		$bool = 0;
	}
	return( $bool );
}

###############################################################################
# Subroutine:  is_val_range()
#
#   Parameters:
#		$x = a reference to an array.
#		$min = the minimum value allowed in the array.
#		$max = the maximum value allowed in the array.
#
#   Usage:
#		( &is_val_range( \@list, $min, $max ) );
#
#   Output:
#		This is a boolean value return, basically.  The subroutine returns 1,
#		or true, if the passed in array contains only integers that are >= $min
#		and <= $max, 0, or false, otherwise.
###############################################################################

sub is_val_range {
	my( $x, $min, $max ) = @_;
	my( @temp ) = @$x;
	my( $y, $bool );
	$bool = 1;
	foreach $y (@temp) {
		if( ($y < $min) || ($y > $max) || ($y !~ /^-?\d+$/o) ) {
			$bool = 0;
			last;
		}
	}
	return( $bool );
}

return( 1 );